import Chart from 'chart.js'

import results from '../results/index.js'

function processHash() {
  if (window.location.hash) {
    let set = window.location.hash.substring(1)

    if (!results.hasOwnProperty(set)) {
      console.error(`No result set "${set}" defined`)
      return
    }

    setCharts(set)
    setDimensions(set)
    printStats()
  }
}

// Dimension rendering

const dimensionMap = {
  set1: [[0], [0, 1, 2], [0], [0, 1], [0]],
  set2: [[0], [2], [0], [0, 1], [1]],
  set3: [[0], [0, 3], [1], [0, 1], [0]],
  set4: [[0], [0], [1], [0, 1], [0, 2]],
  set5: [[1], [0, 3], [0], [0, 1], [0]]
}
function setDimensions(set) {
  let map = dimensionMap[set]
  let dimensions = document.getElementsByClassName('characteristics')

  for (let i = 0; i < dimensions.length; i++) {
    let dim = dimensions[i]
    for (let j = 0; j < dim.children.length; j++) {
      let child = dim.children[j]
      if (map[i].indexOf(j) !== -1) child.setAttribute('active', 'active')
      else child.removeAttribute('active')
    }
  }
}

// Chart rendering

const charts = []

function killChildren(node) {
  while (node.hasChildNodes()) node.removeChild(node.lastChild)
}

function createChart(config) {
  let canvas = document.createElement('canvas')
  let ctx = canvas.getContext('2d')
  charts.push(new Chart(ctx, config))
  return canvas
}

function destroyCharts() {
  while (charts.length) {
    let chart = charts.pop()
    chart.destroy()
  }
}

function setCharts(set) {
  let resultsRoot = document.getElementById('chartRoot')
  destroyCharts()
  killChildren(resultsRoot)
  for (let config of results[set]) {
    resultsRoot.appendChild(createChart(config))
  }
}

// toggle button

function togglePerformanceMeasure() {
  const toggleLabel = label =>
    label === 'Performance - Milliseconds per operation (ms)'
      ? 'Performance - Operations per second (hz)'
      : 'Performance - Milliseconds per operation (ms)'

  for (let chart of charts) {
    let datasets = chart.config.data.datasets
    for (let dataset of datasets) {
      if (dataset.id !== 'numberOfNodes') {
        dataset.data = dataset.data.map(val => 1000 / val)
      }
    }
    let scaleLabel = chart.config.options.scales.yAxes[0].scaleLabel
    scaleLabel.labelString = toggleLabel(scaleLabel.labelString)
    chart.update()
  }
  printStats()
}

// stats

function printStats() {
  const sum = arr => arr.reduce((a, b) => a + b, 0)
  const mean = arr => sum(arr) / arr.length
  const median = arr => [...arr].sort()[Math.floor(arr.length / 2)]
  let i = 0
  for (let chart of charts) {
    console.log(`=== Stats for chart #${i} ===`)
    let datasets = chart.config.data.datasets
    for (let dataset of datasets) {
      if (dataset.data.length === 0 || dataset.id === 'numberOfNodes') continue
      let data = dataset.data
      console.log(
        `Dataset: ${dataset.label}, mean: ${mean(data)}, median: ${median(
          data
        )}`
      )
    }
    i++
  }
}

// init

processHash()
window.onhashchange = processHash
document
  .getElementById('toggle')
  .addEventListener('click', togglePerformanceMeasure)
